from __future__ import annotations

from digitalhub.entities.dataitem._base.spec import DataitemSpec, DataitemValidator


class DataitemSpecIceberg(DataitemSpec):
    """
    DataitemSpecIceberg specifications.
    """


class DataitemValidatorIceberg(DataitemValidator):
    """
    DataitemValidatorIceberg validator.
    """
