from __future__ import annotations

from digitalhub.entities._base.entity.spec import Spec, SpecValidator


class FunctionSpec(Spec):
    """
    FunctionSpec specifications.
    """


class FunctionValidator(SpecValidator):
    """
    FunctionValidator validator.
    """
