from __future__ import annotations

from pydantic import BaseModel


class SqlStoreConfig(BaseModel):
    """
    SQL store configuration class.
    """

    host: str
    """SQL host."""

    port: int
    """SQL port."""

    user: str
    """SQL user."""

    password: str
    """SQL password."""

    database: str
    """SQL database name."""
