from __future__ import annotations

from digitalhub.entities._base.material.spec import MaterialSpec, MaterialValidator


class DataitemSpec(MaterialSpec):
    """
    DataitemSpec specifications.
    """


class DataitemValidator(MaterialValidator):
    """
    DataitemValidator validator.
    """
