from __future__ import annotations

from digitalhub.entities._base.material.status import MaterialStatus


class ArtifactStatus(MaterialStatus):
    """
    Status class for artifact entities.
    """


class ArtifactStatusArtifact(ArtifactStatus):
    """
    Status class for artifact entities.
    """
