from __future__ import annotations

from digitalhub.entities.dataitem._base.status import DataitemStatus


class DataitemStatusDataitem(DataitemStatus):
    """
    DataitemStatusDataitem status.
    """
